#ifndef DISASM_H
#define DISASM_H

#pragma pack(push)
#pragma pack(1)

namespace ade {

  typedef unsigned char   Byte;
  typedef unsigned short  Word;
  typedef unsigned long   Long;

  static const char* errors[]= {
    "No errors",
    "Out of memory",
    "Need addition memory",
    "Unknown instrution",
    "Illegal instruction",
    "Invalid operand",
    "Extra characters on line",
    "Invalid address",
    "Value out of range",
    "Unassembled instruction cross end of code block",
    "Operand sizes do not match",
    "Superfluous prefix"
  };

  enum Errors {
    e86_no_errors,
    // Memory limit too low
   	e86_out_of_memory,
    // need addition memory to produce code
    e86_addition_memory,
    // unknown instruction
    e86_unknown_instruction,
    // instruction not found
    e86_illegal_instruction,
    // invalid instruction operand
    e86_invalid_operand,
    // unexpected characters on line
    e86_extra_characters,
    // address expression invalid
    e86_invalid_address,
    // value out of range
    e86_out_of_range,
    // unassembled instruction cross end of code block
    e86_cross_memory,
    // operands size not match
    e86_size_not_match,
    // superfluous prefix
    e86_superfluous_prefix
  };


  /* Instruction encoding flags */

  enum Flags {
    C_ADDR1=0x0000001,      // Address bytes 1
    C_ADDR2=0x0000002,      // Address bytes 2
    C_ADDRX=0x0000004,      // Address bytes 2 or 4
    C_REL  =0x0000008,      // Relative address
    C_DATA1=0x0000010,      // Data bytes 1
    C_DATA2=0x0000020,      // Data bytes 2
    C_DATAX=0x0000040,      // Data bytes 2 or 4
    C_ENTER=0x0000030,      // "ENTER" instruction data (Data 1 | Data 2)
    C_BITS =0x0000080,      // Instruction have bit S (signed value)
    C_BITW =0x0000100,      // Instruction have bit W (force 8 bit size)
    C_BITD =0x0000200,      // Instruction have bit D (operand exchagne)
    C_TBL2 =0x0000400,      // Instruction use addition table (0F opcode)
    C_CC   =0x0000800,      // Instruction have condition code
    C_AS16 =0x0001000,      // Default 16 bit address
    C_DS16 =0x0002000,      // Default 16 bit data
    C_USE16=0x0004000,      // Default mode 16 bit
    C_SIZEQ=0x0008000,      // Operands size equalent (first = second)
    C_MODRM=0x0010000,      // Instruction have MODRM byte
    C_MODEX=0x0020000,      // Opcode extension in MODRM byte
    C_SIB  =0x0040000,      // Instruction have SIB byte
    C_COND =0x0080000,      // Have condition
    C_JMP  =0x0100000,      // JMP instruction
    C_CALL =0x0200000,      // CALL instruciton
    C_RET  =0x0400000,      // RET instruction
    C_FAR  =0x0800000,      // Far JMP/CALL
    C_SFMT =0x1000000       // print address part as '%s'
  };


   /*******************
   * ARG86 structure *
   *******************/

  typedef struct {
  /* 00*/ Byte    spec;           // Type specified field
  /* 01*/ Byte    optype;         // Operand type
  } ARG86;

  enum ArgLocation {
    L_RG   =0x10,           // Opcode filed
    L_MRG  =0x20,           // ModReg field
    L_MRM  =0x40,           // ModRegMem field
    L_SEG  =0x50            // Opcode filed 000xx000B
  };

  enum ArgType {
    T_REG  =0x10,           // General register
    T_CRX  =0x20,           // CRX register
    T_DRX  =0x30,           // DRX register
    T_STX  =0x40,           // STX register
    T_MMX  =0x50,           // MMX register
    T_XMM  =0x60,           // XMM register
    T_SEG  =0x70,           // Segment register
    T_MEM  =0x80,           // Memory pointer
    T_RM   =0x90,           // Register or Memory
    T_REL  =0xA0,           // Relative address
    T_OFFS =0xB0,           // Direct offset
    T_FAR  =0xC0,           // Far address
    T_FPTR =0xD0,           // Far memory poiner
    T_IMM  =0xE0,           // Immediate value
    T_TRX  =0xF0            // TRX register
  };

  enum AddrType {
    S_SCALE=0x01,           // Scale factor
    S_INDEX=0x02,           // Index register
    S_BASE =0x04,           // Base register
    S_DISP =0x08,           // Displacement
    S_AM16 =0x10,           // Address mode 16
    S_SHORT=0x20,           // Short address
    S_SIGND=0x08            // Signed value
  };


  /*******************
   * TBL86 structure *
   *******************/

  struct TBL86 {
  /* 00*/ Byte    next;           // Offset to next item or
  /* 01*/ Byte    opcode;         // Base opcode
  /* 02*/ Byte    opmask;         // Opcode search mask
  /* 03*/ Byte    grpmf;          // Group index and instruction mode flags
  /* 04*/ Long    flags;          // Instruction encoding flags
  /* 08*/ Word    op1;            // Instruction operands
  /* 0A*/ Word    op2;
  /* 0C*/ Word    op3;
  /* 0E*/ Byte    name[1];        // Name of instruction (variable length array)
  };


  /*******************
   * DIS86 structure *
   *******************/

  struct DIS86 {
  /* 00*/ Long    error;          // Error detected or 0 if no errors
  /* 04*/ Long    ip;             // Instruction pointer
  /* 08*/ Byte    size;           // Size of instruction
  /* 09*/ Byte    p_66;           // Data override prefix
  /* 0A*/ Byte    p_67;           // Address override prefix
  /* 0B*/ Byte    p_rep;          // Repeat prefix
  /* 0C*/ Byte    p_seg;          // Segment override prefix
  /* 0D*/ Byte    p_lock;         // Bus lock prefix
  /* 0E*/ Byte    byte0f;         // Previous opcode byte 0x0F
  /* 0F*/ Byte    opcode;         // Opcode byte
  /* 10*/ Byte    modrm;          // MODRM byte
  /* 11*/ Byte    sib;            // SIB byte
  /* 12*/ Long    address;        // Address
  /* 16*/ Word    selector;       // Selector
  /* 18*/ Long    immval;         // Immediate constant
  /* 1C*/ Long    flags;          // Instruction flags
  /* 20*/ TBL86*  tbl_entry;      // Instruction decoding information
  /* 24*/ Word    op1;            // Instruction operands
  /* 26*/ Word    op2;
  /* 28*/ Word    op3;
  };


  typedef struct {
  } DBG86;


  extern "C"
  {
  unsigned int __cdecl
  assemble(Byte* memptr, Byte* maxptr, const char* source, Long org, int bits);

  unsigned int __cdecl
  unassemble(void* binptr, void* maxptr, struct DIS86* result, Long org, int bits);

  unsigned int __cdecl
  print_dis86(void* outptr, void* maxptr, struct DIS86* disasm, DBG86* debug);
  }

}

#pragma pack(pop)

#endif//DISASM_H
